"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const DeviceSlot_1 = require("./model/DeviceSlot");
const Slot_1 = require("./model/Slot");
const three_1 = require("three");
const aframe_typescript_boilerplate_1 = require("aframe-typescript-boilerplate");
const InteractionMode_1 = require("./model/InteractionMode");
class InterfaceSystemController extends aframe_typescript_boilerplate_1.AbstractSystemController {
    constructor(system, scene, data) {
        super(system, scene, data);
        this.interactionMode = InteractionMode_1.InteractionMode.FP;
        this.uiFocus = false;
        this.selfScale = 1;
        this.collidables = new Array();
        this.devices = new Map();
        this.tools = new Map();
        this.toolNames = new Array();
        this.slots = new Map();
        this.slotListeners = new Map();
        this.cameraPosition = new three_1.Vector3(0, 0, 0);
        this.cameraDirection = new three_1.Vector3(0, 0, 0);
        this.interfaceEntity = this.scene.querySelector('[interface]');
        if (this.interfaceEntity) {
            console.info("interface entity set.");
        }
        else {
            throw new Error("interface entity not found.");
        }
        this.cameraEntity = this.interfaceEntity.querySelector('[camera]');
        if (this.cameraEntity) {
            console.info("interface camera entity set.");
        }
        else {
            throw new Error("interface camera entity not found.");
        }
    }
    init() {
        const materialUiSystem = this.getSystemController("material-ui");
        materialUiSystem.onFocus = () => {
            this.uiFocus = true;
            const lookControls = this.cameraEntity.components["look-controls"];
            if (lookControls) {
                lookControls.pause();
            }
        };
        materialUiSystem.onFocusOut = () => {
            this.uiFocus = false;
            const lookControls = this.cameraEntity.components["look-controls"];
            if (lookControls) {
                lookControls.play();
            }
        };
    }
    pause() {
    }
    play() {
    }
    tick(time, timeDelta) {
        this.cameraPosition.copy(this.cameraEntity.object3D.position);
        this.cameraPosition = this.cameraEntity.object3D.parent.localToWorld(this.cameraPosition);
        this.cameraEntity.object3D.getWorldDirection(this.cameraDirection);
        this.cameraDirection.multiplyScalar(-1);
    }
    getCollidables() {
        return this.collidables;
    }
    addCollidable(object) {
        this.collidables.push(object);
    }
    removeCollidable(object) {
        const index = this.collidables.indexOf(object, 0);
        if (index > -1) {
            this.collidables.splice(index, 1);
        }
    }
    setInterfaceController(interfaceController) {
        this.interfaceController = interfaceController;
    }
    setDevice(slot, device) {
        if (this.devices.has(slot)) {
            console.warn("interface already has controls at: " + DeviceSlot_1.DeviceSlot[slot]);
        }
        else {
            this.devices.set(slot, device);
            //console.log("interface controls " + device.componentName + " set at: " + DeviceSlot[slot]);
        }
    }
    getDevice(slot) {
        return this.devices.get(slot);
    }
    registerTool(tool) {
        if (!this.tools.has(tool.componentName)) {
            //console.log("interface tool '" + tool.componentName + "' registered.");
            this.tools.set(tool.componentName, tool);
            this.toolNames.push(tool.componentName);
        }
    }
    registerSlotListener(slot, slotListener) {
        if (!this.slotListeners.has(slot)) {
            this.slotListeners.set(slot, new Array());
        }
        this.slotListeners.get(slot).push(slotListener);
        // Notify slot listener of slotted tool if tool already slotted.
        const slottedTool = this.getToolAtSlot(Slot_1.Slot.PRIMARY);
        if (slottedTool) {
            slotListener.onToolSlotted(Slot_1.Slot.PRIMARY, slottedTool.componentName);
        }
    }
    getTool(name) {
        if (this.tools.has(name)) {
            return this.tools.get(name);
        }
        else {
            throw new Error("Tool '" + name + "' not registered.");
        }
    }
    slotTool(slot, tool) {
        this.slots.set(slot, tool);
        //console.log("interface tool " + tool.componentName + " set at: " + Slot[slot]);
        if (this.slotListeners.has(slot)) {
            this.slotListeners.get(slot).forEach((slotListener) => {
                slotListener.onToolSlotted(slot, tool.componentName);
            });
        }
    }
    getToolAtSlot(slot) {
        return this.slots.get(slot);
    }
    buttonUp(device, slot, button) {
        if (this.slots.has(slot)) {
            this.slots.get(slot).buttonUp(device, slot, button);
        }
    }
    buttonDown(device, slot, button) {
        if (this.slots.has(slot)) {
            this.slots.get(slot).buttonDown(device, slot, button);
        }
    }
    stickTwist(device, slot, stick, x, y) {
        if (this.slots.has(slot)) {
            this.slots.get(slot).stickTwist(device, slot, stick, x, y);
        }
    }
    slotNextTool(slot) {
        if (this.slots.has(slot)) {
            const tool = this.slots.get(slot);
            const toolIndex = this.toolNames.indexOf(tool.componentName);
            const nextToolIndex = (toolIndex == this.toolNames.length - 1) ? 0 : toolIndex + 1;
            //console.log("tool index: " + toolIndex + " next tool index: " + nextToolIndex + " lenght: " + this.toolNames.length);
            //console.log(this.toolNames);
            const nextToolName = this.toolNames[nextToolIndex];
            const nextTool = this.getTool(nextToolName);
            this.slotTool(slot, nextTool);
        }
    }
    slotPreviousTool(slot) {
        if (this.slots.has(slot)) {
            const tool = this.slots.get(slot);
            const toolIndex = this.toolNames.indexOf(tool.componentName);
            const previousToolIndex = (toolIndex == 0) ? this.toolNames.length - 1 : toolIndex - 1;
            const previousToolName = this.toolNames[previousToolIndex];
            const previousTool = this.getTool(previousToolName);
            this.slotTool(slot, previousTool);
        }
    }
    setSelfScale(selfScale) {
        const scale = this.interfaceEntity.getAttribute("scale");
        this.selfScale = selfScale;
        scale.x = this.selfScale;
        scale.y = this.selfScale;
        scale.z = this.selfScale;
        const spaceSystem = this.getSystemController("space");
        spaceSystem.sendAvatarDescriptionUpdate();
    }
    getSelfScale() {
        return this.selfScale;
    }
    isVrMode() {
        return this.getToolAtSlot(Slot_1.Slot.PRIMARY_SELECTOR).vrMode;
    }
    isUiFocus() {
        return this.uiFocus;
    }
    setVrMode(vrMode) {
        if (this.isVrMode() === vrMode) {
            return;
        }
        else {
            if (vrMode) {
                this.scene.enterVR();
            }
            else {
                this.scene.exitVR();
            }
        }
    }
    getInteractionMode() {
        return this.interactionMode;
    }
    setInteractionMode(interactionMode) {
        if (interactionMode === this.interactionMode) {
            return;
        }
        this.interactionMode = interactionMode;
    }
}
InterfaceSystemController.DEFINITION = new aframe_typescript_boilerplate_1.SystemControllerDefinition("interface", {}, (system, scene, data) => new InterfaceSystemController(system, scene, data));
exports.InterfaceSystemController = InterfaceSystemController;
//# sourceMappingURL=InterfaceSystemController.js.map