"use strict";
/**
 * @author mrdoob, tlaukkan (copied from threejs BufferGeometryUtil.js typescript and added extendability, dropped morph attributes support)
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const three_1 = require("three");
class BufferGeometryMerge {
    constructor(material) {
        this.groupOffset = 0;
        this.indexOffset = 0;
        this.indexMerge = new Array();
        this.attributeMerges = new Map();
        this.geometry = new three_1.BufferGeometry();
        this.freeObjectIndexes = new Set();
        this.material = material;
    }
}
exports.BufferGeometryMerge = BufferGeometryMerge;
class BufferAttributeMerge {
    constructor(attribute, offset) {
        this.objectIndexAttributeRanges = new Map();
        this.offset = offset;
        this.attribute = attribute;
    }
}
function mergeBufferGeometries(merge, geometries, useGroups) {
    return __awaiter(this, void 0, void 0, function* () {
        const mergedGeometry = merge.geometry;
        if (geometries.length > 0) {
            const isIndexed = geometries[0].index !== null;
            const attributeNames = new Set(Object.keys(geometries[0].attributes));
            const attributes = {};
            for (let i = 0; i < geometries.length; ++i) {
                const geometry = geometries[i];
                // ensure that all geometries are indexed, or none
                if (isIndexed !== (geometry.index !== null))
                    throw new Error("Inconsistent indexed in merged geometries.");
                // gather attributes, exit early if they're different
                for (const name in geometry.attributes) {
                    if (!attributeNames.has(name))
                        throw new Error("Inconsistent attributes in merged geometries.");
                    if (attributes[name] === undefined)
                        attributes[name] = [];
                    if (geometry.mergeObjectIndex === undefined) {
                        throw new Error("Merge object index not defined for merging geometry.");
                    }
                    geometry.attributes[name].mergeObjectIndex = geometry.mergeObjectIndex;
                    attributes[name].push(geometry.attributes[name]);
                }
                mergedGeometry.userData.mergedUserData = mergedGeometry.userData.mergedUserData || [];
                mergedGeometry.userData.mergedUserData.push(mergedGeometry.userData);
                if (useGroups) {
                    let count;
                    if (isIndexed) {
                        count = geometry.index.count;
                    }
                    else if (geometry.attributes.position !== undefined) {
                        count = geometry.attributes.position.count;
                    }
                    else {
                        throw new Error("Groups used in geometry merging but group count could not be resolved");
                    }
                    mergedGeometry.addGroup(merge.groupOffset, count, i);
                    merge.groupOffset += count;
                }
            }
            // merge indices
            if (isIndexed) {
                for (let i = 0; i < geometries.length; ++i) {
                    const index = geometries[i].index;
                    for (let j = 0; j < index.count; ++j) {
                        merge.indexMerge.push(index.getX(j) + merge.indexOffset);
                    }
                    merge.indexOffset += geometries[i].attributes.position.count;
                }
                mergedGeometry.setIndex(merge.indexMerge);
            }
            for (const name in attributes) {
                //console.log(name);
                const arrayLength = sumArrayLength(attributes[name]);
                if (!merge.attributeMerges.has(name)) {
                    // Lets add build capacity for 1000 more geometries.
                    const extendedArrayLength = arrayLength + attributes[name][0].array.length * 1000;
                    const attributeMerge = createBufferAttributeMerge(attributes[name], extendedArrayLength);
                    merge.attributeMerges.set(name, attributeMerge);
                }
                else {
                    const attributeMerge = merge.attributeMerges.get(name);
                    if (attributeMerge.offset + arrayLength >= attributeMerge.attribute.array.length) {
                        throw new Error("merge out of capacity");
                    }
                }
            }
            for (const name in attributes) {
                const mergedAttribute = yield mergeBufferAttributes(merge.attributeMerges.get(name), name, attributes[name]);
                if (!mergedGeometry.attributes[name]) {
                    mergedGeometry.addAttribute(name, mergedAttribute.attribute);
                }
            }
        }
        for (const attributeMerge of merge.attributeMerges.values()) {
            attributeMerge.attribute.needsUpdate = true;
        }
        mergedGeometry.computeBoundingSphere();
        mergedGeometry.computeVertexNormals();
        return mergedGeometry;
    });
}
exports.mergeBufferGeometries = mergeBufferGeometries;
function createBufferAttributeMerge(attributes, arrayLength) {
    const array = new attributes[0].array.constructor(arrayLength);
    const merge = new BufferAttributeMerge(new three_1.BufferAttribute(array, attributes[0].itemSize, attributes[0].normalized), 0);
    // Zero as no attributes added yet.
    merge.attribute.count = 0;
    return merge;
}
function sumArrayLength(attributes) {
    let arrayLength = 0;
    for (let i = 0; i < attributes.length; ++i) {
        arrayLength += attributes[i].array.length;
    }
    return arrayLength;
}
function mergeBufferAttributes(merge, name, attributes) {
    return __awaiter(this, void 0, void 0, function* () {
        for (const attribute of attributes) {
            if (attribute.isInterleavedBufferAttribute)
                throw new Error("Attributes had interleaved attributes..");
            yield mergeBufferAttributeWithDelay(merge, attribute);
        }
        return merge;
    });
}
let niceTimeoutCounter = 0;
function mergeBufferAttributeWithDelay(merge, attribute) {
    return new Promise(function (resolve, reject) {
        mergeBufferAttribute(merge, attribute).then(() => {
            niceTimeoutCounter++;
            if (niceTimeoutCounter % 100 == 0) {
                setTimeout(() => {
                    resolve();
                }, 1);
            }
            else {
                resolve();
            }
        }).catch((error) => {
            reject(error);
        });
    });
}
function mergeBufferAttribute(merge, attribute) {
    return __awaiter(this, void 0, void 0, function* () {
        if (merge.attribute.itemSize !== attribute.itemSize)
            throw new Error("Inconsistent item size in merged attributes.");
        if (merge.attribute.normalized !== attribute.normalized)
            throw new Error("Inconsistent normalized in merged attributes.");
        const mergeObjectIndex = attribute.mergeObjectIndex;
        if (mergeObjectIndex === undefined) {
            throw new Error("Merge object index not defined for merging attribute.");
        }
        if (!merge.objectIndexAttributeRanges.has(mergeObjectIndex)) {
            merge.objectIndexAttributeRanges.set(mergeObjectIndex, []);
        }
        merge.objectIndexAttributeRanges.get(mergeObjectIndex).push([merge.offset, attribute.array.length]);
        merge.attribute.array.set(attribute.array, merge.offset);
        merge.attribute.count = merge.attribute.count + attribute.array.length / attribute.itemSize;
        merge.offset += attribute.array.length;
    });
}
function clearBufferGeometries(merge, geometries) {
    const mergedGeometry = merge.geometry;
    if (geometries.length > 0) {
        const attributes = {};
        for (let i = 0; i < geometries.length; ++i) {
            const geometry = geometries[i];
            for (const name in geometry.attributes) {
                if (attributes[name] === undefined)
                    attributes[name] = [];
                if (geometry.mergeObjectIndex === undefined) {
                    throw new Error("Merge object index not defined for merging geometry.");
                }
                // Store merge object index to free object indexes.
                if (!merge.freeObjectIndexes.has(geometry.mergeObjectIndex)) {
                    merge.freeObjectIndexes.add(geometry.mergeObjectIndex);
                }
                geometry.attributes[name].mergeObjectIndex = geometry.mergeObjectIndex;
                attributes[name].push(geometry.attributes[name]);
            }
        }
        for (const name in attributes) {
            if (name == "position") { // clear only positions to disable the geometry
                clearBufferAttributes(merge.attributeMerges.get(name), name, attributes[name]);
            }
        }
    }
    for (const attributeMerge of merge.attributeMerges.values()) {
        attributeMerge.attribute.needsUpdate = true;
    }
    return mergedGeometry;
}
exports.clearBufferGeometries = clearBufferGeometries;
function clearBufferAttributes(merge, name, attributes) {
    for (const attribute of attributes) {
        clearBufferAttribute(merge, attribute);
    }
}
function clearBufferAttribute(merge, attribute) {
    if (merge.attribute.itemSize !== attribute.itemSize)
        throw new Error("Inconsistent item size in merged attributes.");
    if (merge.attribute.normalized !== attribute.normalized)
        throw new Error("Inconsistent normalized in merged attributes.");
    const mergeObjectIndex = attribute.mergeObjectIndex;
    if (mergeObjectIndex === undefined) {
        throw new Error("Merge object index not defined for merging attribute.");
    }
    if (!merge.objectIndexAttributeRanges.has(mergeObjectIndex)) {
        throw new Error("Merge object index does not have attribute ranges to clear.");
    }
    for (const range of merge.objectIndexAttributeRanges.get(mergeObjectIndex)) {
        const offset = range[0];
        const length = range[1];
        for (let i = 0; i < length; i++) {
            merge.attribute.array[offset + i] = 0;
        }
    }
}
function updateBufferGeometries(merge, geometries) {
    const mergedGeometry = merge.geometry;
    if (geometries.length > 0) {
        const attributes = {};
        for (let i = 0; i < geometries.length; ++i) {
            const geometry = geometries[i];
            for (const name in geometry.attributes) {
                if (attributes[name] === undefined)
                    attributes[name] = [];
                if (geometry.mergeObjectIndex === undefined) {
                    throw new Error("Merge object index not defined for merging geometry.");
                }
                geometry.attributes[name].mergeObjectIndex = geometry.mergeObjectIndex;
                attributes[name].push(geometry.attributes[name]);
            }
        }
        for (const name in attributes) {
            if (name == "position") { // update only positions to disable the geometry
                updateBufferAttributes(merge.attributeMerges.get(name), name, attributes[name]);
            }
        }
    }
    for (const attributeMerge of merge.attributeMerges.values()) {
        attributeMerge.attribute.needsUpdate = true;
    }
    mergedGeometry.computeBoundingSphere();
    mergedGeometry.computeVertexNormals();
    return mergedGeometry;
}
exports.updateBufferGeometries = updateBufferGeometries;
function updateBufferAttributes(merge, name, attributes) {
    for (const attribute of attributes) {
        updateBufferAttribute(merge, attribute);
    }
}
function updateBufferAttribute(merge, attribute) {
    if (merge.attribute.itemSize !== attribute.itemSize)
        throw new Error("Inconsistent item size in merged attributes.");
    if (merge.attribute.normalized !== attribute.normalized)
        throw new Error("Inconsistent normalized in merged attributes.");
    const mergeObjectIndex = attribute.mergeObjectIndex;
    if (mergeObjectIndex === undefined) {
        throw new Error("Merge object index not defined for merging attribute.");
    }
    if (!merge.objectIndexAttributeRanges.has(mergeObjectIndex)) {
        throw new Error("Merge object index does not have attribute ranges to clear.");
    }
    for (const range of merge.objectIndexAttributeRanges.get(mergeObjectIndex)) {
        const offset = range[0];
        const length = range[1];
        for (let i = 0; i < length; i++) {
            merge.attribute.array[offset + i] = attribute.array[i];
        }
    }
}
//# sourceMappingURL=geometry_merge_util.js.map