"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const three_1 = require("three");
const geometry_merge_util_1 = require("./geometry_merge_util");
class ObjectMerge {
    constructor() {
        this.group = new three_1.Group();
        this.geometryMerges = new Map();
        this.objectOffset = 0;
    }
}
exports.ObjectMerge = ObjectMerge;
function mergeObject3Ds(merge, objects) {
    return __awaiter(this, void 0, void 0, function* () {
        const geometryDataMap = new Map();
        for (const object of objects) {
            yield collectBufferGeometries(merge, object, geometryDataMap);
        }
        const allGeometryIds = new Set();
        for (const geometryId of geometryDataMap.keys()) {
            allGeometryIds.add(geometryId);
        }
        for (const geometryId of merge.geometryMerges.keys()) {
            allGeometryIds.add(geometryId);
        }
        for (const geometryId of allGeometryIds) {
            const geometries = new Array();
            const geometryDataArray = geometryDataMap.get(geometryId);
            if (geometryDataArray) {
                //console.log(geometryId + ": " + geometryDataMap.get(geometryId)!!.length);
                const material = geometryDataArray[0].material;
                for (const geometryData of geometryDataArray) {
                    geometries.push(geometryData.geometry);
                }
                if (!merge.geometryMerges.has(geometryId)) {
                    //console.log("adding geometry merge: " + geometryId);
                    const geometryMerge = new geometry_merge_util_1.BufferGeometryMerge(material);
                    merge.geometryMerges.set(geometryId, geometryMerge);
                    const mesh = new three_1.Mesh(geometryMerge.geometry, geometryMerge.material);
                    mesh.castShadow = true;
                    mesh.receiveShadow = true;
                    merge.group.add(mesh);
                }
            }
            const geometryMerge = merge.geometryMerges.get(geometryId);
            yield geometry_merge_util_1.mergeBufferGeometries(geometryMerge, geometries, false);
        }
    });
}
exports.mergeObject3Ds = mergeObject3Ds;
function clearObject3Ds(merge, objects) {
    return __awaiter(this, void 0, void 0, function* () {
        const geometryDataMap = new Map();
        // TODO Optimize this for clearing so that geometries are not cloned.
        for (const object of objects) {
            yield collectBufferGeometries(merge, object, geometryDataMap);
        }
        for (const geometryId of geometryDataMap.keys()) {
            const geometries = new Array();
            const geometryDataArray = geometryDataMap.get(geometryId);
            if (geometryDataArray) {
                for (const geometryData of geometryDataArray) {
                    geometries.push(geometryData.geometry);
                }
            }
            geometry_merge_util_1.clearBufferGeometries(merge.geometryMerges.get(geometryId), geometries);
        }
    });
}
exports.clearObject3Ds = clearObject3Ds;
function updateObject3Ds(merge, objects) {
    return __awaiter(this, void 0, void 0, function* () {
        const geometryDataMap = new Map();
        for (const object of objects) {
            yield collectBufferGeometries(merge, object, geometryDataMap);
        }
        for (const geometryId of geometryDataMap.keys()) {
            const geometries = new Array();
            const geometryDataArray = geometryDataMap.get(geometryId);
            if (geometryDataArray) {
                for (const geometryData of geometryDataArray) {
                    geometries.push(geometryData.geometry);
                }
            }
            geometry_merge_util_1.updateBufferGeometries(merge.geometryMerges.get(geometryId), geometries);
        }
    });
}
exports.updateObject3Ds = updateObject3Ds;
class GeometryData {
    constructor(geometry, material) {
        this.geometry = geometry;
        this.material = material;
    }
}
function collectChildBufferGeometries(merge, children, geometries) {
    return __awaiter(this, void 0, void 0, function* () {
        for (const child of children) {
            yield collectBufferGeometriesWithDelay(merge, child, geometries);
        }
    });
}
let niceTimeoutCounter = 0;
function collectBufferGeometriesWithDelay(merge, object, geometries) {
    return new Promise(function (resolve, reject) {
        collectBufferGeometries(merge, object, geometries).then(() => {
            niceTimeoutCounter++;
            if (niceTimeoutCounter % 50 == 0) {
                setTimeout(() => {
                    resolve();
                }, 1);
            }
            else {
                resolve();
            }
        }).catch((error) => {
            reject(error);
        });
    });
}
function collectBufferGeometries(merge, object, geometries) {
    return __awaiter(this, void 0, void 0, function* () {
        //console.log("collecting buffer geometries...");
        object.updateMatrix();
        object.updateMatrixWorld(true);
        //console.log(child.type);
        if (object.type == "Mesh") {
            const mesh = object;
            const geometryUuid = mesh.geometry.uuid;
            const geometry = mesh.geometry.clone();
            if (object.userData.mergeObjectIndex === undefined) {
                throw new Error("Merge object index of merging object is undefined.");
            }
            geometry.mergeObjectIndex = object.userData.mergeObjectIndex;
            if (geometry.type == "BufferGeometry") {
                const bufferGeometry = geometry;
                const material = mesh.material;
                geometry.applyMatrix(mesh.matrixWorld);
                if (!geometries.has(geometryUuid)) {
                    geometries.set(geometryUuid, []);
                }
                geometries.get(geometryUuid).push(new GeometryData(bufferGeometry, material));
            }
        }
        if (object.children && object.children.length > 0) {
            yield collectChildBufferGeometries(merge, object.children, geometries);
        }
    });
}
function cloneObject3D(source) {
    const target = cloneObject3DCore(source);
    target.name = source.name;
    target.up.copy(source.up);
    target.position.copy(source.position);
    target.quaternion.copy(source.quaternion);
    target.scale.copy(source.scale);
    target.matrix.copy(source.matrix);
    target.matrixWorld.copy(source.matrixWorld);
    target.matrixAutoUpdate = source.matrixAutoUpdate;
    target.matrixWorldNeedsUpdate = source.matrixWorldNeedsUpdate;
    target.layers.mask = source.layers.mask;
    target.visible = source.visible;
    target.castShadow = source.castShadow;
    target.receiveShadow = source.receiveShadow;
    target.frustumCulled = source.frustumCulled;
    target.renderOrder = source.renderOrder;
    target.userData = JSON.parse(JSON.stringify(source.userData));
    for (let i = 0; i < source.children.length; i++) {
        const child = source.children[i];
        target.add(cloneObject3D(child));
    }
    return target;
}
exports.cloneObject3D = cloneObject3D;
function cloneObject3DCore(source) {
    if (source.type == "Mesh") {
        const target = new three_1.Mesh(source.geometry, source.material);
        three_1.Object3D.prototype.clone.call(source, target);
        return target;
    }
    else {
        return new source.constructor();
    }
}
//# sourceMappingURL=merge_util.js.map